import os
from datetime import datetime, timedelta
import pandas as pd
import feather
from nemosis import filters, downloader, processing_info_maps, defaults, custom_tables
print(custom_tables.__file__)


def dynamic_data_compiler(start_time, end_time, table_name, raw_data_location, select_columns=None, filter_cols=None,
                          filter_values=None):
    print('Compiling data for table {}.'.format(table_name))
    # Generic setup common to all tables.
    if select_columns is None:
        select_columns = defaults.table_columns[table_name]

    # Pre loop setup, done at table type basis.
    date_filter = processing_info_maps.filter[table_name]
    setup_function = processing_info_maps.setup[table_name]

    if setup_function is not None:
        start_time, end_time = setup_function(start_time, end_time)

    search_type = processing_info_maps.search_type[table_name]

    if search_type == 'all':
        start_search = defaults.nem_data_model_start_time
    elif search_type == 'start_to_end':
        start_search = start_time

    start_time = datetime.strptime(start_time, '%Y/%m/%d %H:%M:%S')
    end_time = datetime.strptime(end_time, '%Y/%m/%d %H:%M:%S')
    start_search = datetime.strptime(start_search, '%Y/%m/%d %H:%M:%S')

    data_tables = dynamic_data_fetch_loop(start_search, start_time, end_time, table_name, raw_data_location,
                                          select_columns, date_filter, search_type)

    all_data = pd.concat(data_tables, sort=False)

    finalise_data = processing_info_maps.finalise[table_name]
    if finalise_data is not None:
        for function in finalise_data:
            all_data = function(all_data, start_time, table_name)

    if filter_cols is not None:
        all_data = filters.filter_on_column_value(all_data, filter_cols, filter_values)

    return all_data


def dynamic_data_fetch_loop(start_search, start_time, end_time, table_name, raw_data_location, select_columns,
                             date_filter, search_type):
    data_tables = []
    table_type = defaults.table_types[table_name]
    date_gen = processing_info_maps.date_gen[table_type](start_search, end_time)
    for year, month, day, index in date_gen:
        data = None
        # Write the file names and paths for where the data is stored in the cache.
        filename_full, path_and_name, filename_full_feather, path_and_name_feather = \
            processing_info_maps.write_filename[table_type](table_name, month, year, day, index, raw_data_location)

        # If the data needed is not in the cache then download it.
        if not os.path.isfile(path_and_name):
            if day is None:
                print('Downloading data for table {}, year {}, month {}'.format(table_name, year, month))
            else:
                print('Downloading data for table {}, year {}, month {}, day {}, time {}'.
                      format(table_name, year, month, day, index))

            processing_info_maps.downloader[table_type](year, month, day, index, filename_full, raw_data_location)

        # If the data exists in feather format the read in the data. If it only exists as a csv then read in from the
        # csv and save to feather.
        if os.path.isfile(path_and_name_feather) and os.stat(path_and_name_feather).st_size > 2000:
            data = feather.read_dataframe(path_and_name_feather, select_columns)
        elif os.path.isfile(path_and_name):
            if day is None:
                print('Creating feather file for faster future access of table {}, year {}, month {}.'.
                      format(table_name, year, month))
            else:
                print('Creating feather file for faster future access of table {}, year {}, month {}, day {}, time {}.'.
                      format(table_name, year, month, day, index))
            # Check what headers the data has.
            headers = pd.read_csv(path_and_name, skiprows=[0], nrows=1).columns
            if defaults.table_types[table_name] == 'MMS':
                # Remove columns from the table column list if they are not in the header, this deals with the fact AEMO
                # has added and removed columns over time.
                columns = [column for column in defaults.table_columns[table_name] if column in headers]
                # Read the data from a csv.
                data = pd.read_csv(path_and_name, skiprows=[0], dtype=str, usecols=columns)
                data = data[:-1]
            elif defaults.table_types[table_name] == 'FCAS':
                columns = defaults.table_columns[table_name]
                data = pd.read_csv(path_and_name, skiprows=[0], dtype=str, names=columns)

            # Remove feather files of the same name, deals with case of corrupted files.
            if os.path.isfile(path_and_name_feather):
                os.unlink(path_and_name_feather)
            # Write to feather file.
            data.to_feather(path_and_name_feather)
            # Delete any columns in data that were not explicitly selected.
            if select_columns is not None:
                for column in columns:
                    if column not in select_columns:
                        del data[column]

        if data is not None:
            # Filter by the start and end time.
            if date_filter is not None:
                data = date_filter(data, start_time, end_time)

            data_tables.append(data)

    return data_tables


def static_table(start_time, end_time, table_name, raw_data_location, select_columns=None, filter_cols=None,
                 filter_values=None):
    print('Retrieving static table {}.'.format(table_name))
    path_and_name = raw_data_location + '/' + defaults.names[table_name]
    if not os.path.isfile(path_and_name):
        print('Downloading data for table {}.'.format(table_name))
        downloader.download_csv(defaults.static_table_url[table_name], raw_data_location, path_and_name)

    table = pd.read_csv(raw_data_location + '/' + defaults.names[table_name], dtype=str,
                        names=defaults.table_columns[table_name])
    if select_columns is not None:
        table = table.loc[:, select_columns]
    for column in table.select_dtypes(['object']).columns:
        table[column] = table[column].map(lambda x: x.strip())

    if filter_cols is not None:
        table = filters.filter_on_column_value(table, filter_cols, filter_values)

    return table


def static_table_xl(start_time, end_time, table_name, raw_data_location, select_columns=None, filter_cols=None,
                    filter_values=None):
    path_and_name = raw_data_location + '/' + defaults.names[table_name] + '.xls'
    print('Retrieving static table {}.'.format(table_name))
    if not os.path.isfile(path_and_name):
        print('Downloading data for table {}.'.format(table_name))
        downloader.download_xl(defaults.static_table_url[table_name], raw_data_location, path_and_name)
    xls = pd.ExcelFile(path_and_name)
    table = pd.read_excel(xls, 'Generators and Scheduled Loads', dtype=str)
    table = table.loc[:, select_columns]
    if filter_cols is not None:
        table = filters.filter_on_column_value(table, filter_cols, filter_values)
    table = table.drop_duplicates(['DUID'])

    return table


method_map = {'DISPATCHLOAD': dynamic_data_compiler,
              'DISPATCHPRICE': dynamic_data_compiler,
              'TRADINGLOAD': dynamic_data_compiler,
              'TRADINGPRICE': dynamic_data_compiler,
              'TRADINGREGIONSUM': dynamic_data_compiler,
              'TRADINGINTERCONNECT': dynamic_data_compiler,
              'DISPATCH_UNIT_SCADA': dynamic_data_compiler,
              'DISPATCHCONSTRAINT': dynamic_data_compiler,
              'DUDETAILSUMMARY': dynamic_data_compiler,
              'DUDETAIL': dynamic_data_compiler,
              'GENCONDATA': dynamic_data_compiler,
              'SPDREGIONCONSTRAINT': dynamic_data_compiler,
              'SPDCONNECTIONPOINTCONSTRAINT': dynamic_data_compiler,
              'SPDINTERCONNECTORCONSTRAINT': dynamic_data_compiler,
              'FCAS_4_SECOND': dynamic_data_compiler,
              'ELEMENTS_FCAS_4_SECOND': static_table,
              'VARIABLES_FCAS_4_SECOND': static_table,
              'Generators and Scheduled Loads': static_table_xl,
              'BIDDAYOFFER_D': dynamic_data_compiler,
              'BIDPEROFFER_D': dynamic_data_compiler,
              'FCAS_4s_SCADA_MAP': custom_tables.fcas4s_scada_match,
              'PLANTSTATS': custom_tables.plant_stats,
              'DISPATCHINTERCONNECTORRES': dynamic_data_compiler,
              'DISPATCHREGIONSUM': dynamic_data_compiler,
              'LOSSMODEL': dynamic_data_compiler,
              'LOSSFACTORMODEL': dynamic_data_compiler,
              'MNSP_DAYOFFER': dynamic_data_compiler,
              'MNSP_PEROFFER': dynamic_data_compiler,
              'MNSP_INTERCONNECTOR': dynamic_data_compiler,
              'INTERCONNECTOR': dynamic_data_compiler,
              'INTERCONNECTORCONSTRAINT': dynamic_data_compiler,
              'MARKET_PRICE_THRESHOLDS': dynamic_data_compiler}



import unittest
from datetime import timedelta
from nemosis import data_fetch_methods
from nemosis import defaults
import pandas as pd
from nemosis import custom_tables


class TestDynamicDataCompilerWithSettlementDateFiltering(unittest.TestCase):
    def setUp(self):
        self.table_names = ['BIDDAYOFFER_D', 'BIDPEROFFER_D', 'DISPATCHLOAD', 'DISPATCHCONSTRAINT',
                            'DISPATCH_UNIT_SCADA', 'DISPATCHPRICE', 'DISPATCHINTERCONNECTORRES', 'DISPATCHREGIONSUM',
                            'TRADINGLOAD', 'TRADINGPRICE', 'TRADINGREGIONSUM', 'TRADINGINTERCONNECT']

        self.table_types = {'DISPATCHLOAD': 'DUID', 'DISPATCHCONSTRAINT': 'CONSTRAINTID', 'DISPATCH_UNIT_SCADA': 'DUID',
                       'DISPATCHPRICE': 'REGIONID', 'DISPATCHINTERCONNECTORRES': 'INTERCONNECTORID',
                       'DISPATCHREGIONSUM': 'REGIONID', 'BIDPEROFFER_D': 'DUID-BIDTYPE',
                       'BIDDAYOFFER_D': 'DUID-BIDTYPE', 'TRADINGLOAD': 'DUID', 'TRADINGPRICE': 'REGIONID',
                       'TRADINGREGIONSUM': 'REGIONID', 'TRADINGINTERCONNECT': 'INTERCONNECTORID'}

        self.filter_values = {'DUID': (['AGLHAL'],), 'REGIONID': (['SA1'],), 'INTERCONNECTORID': (['VIC1-NSW1'],),
                              'CONSTRAINTID': (['DATASNAP_DFS_Q_CLST'],), 'DUID-BIDTYPE': (['AGLHAL', 'ENERGY'],)}

    def test_dispatch_tables_start_of_month(self):
        start_time = '2015/05/01 00:00:00'
        end_time = '2015/05/01 05:15:00'
        for table in self.table_names:
            print('Testing {} returing values at start of month.'.format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            cols = [dat_col, self.table_types[table]]
            filter_cols = (self.table_types[table],)
            expected_length = 63
            expected_number_of_columns = 2
            expected_first_time = pd.Timestamp.strptime(start_time, '%Y/%m/%d %H:%M:%S') + timedelta(minutes=5)
            expected_last_time = pd.Timestamp.strptime(end_time, '%Y/%m/%d %H:%M:%S')
            if table in ['TRADINGLOAD', 'TRADINGPRICE', 'TRADINGREGIONSUM', 'TRADINGINTERCONNECT']:
                expected_length = 10
                expected_first_time = '2015/05/01 00:30:00'
                expected_first_time = pd.Timestamp.strptime(expected_first_time, '%Y/%m/%d %H:%M:%S')
                expected_last_time = '2015/05/01 05:00:00'
                expected_last_time = pd.Timestamp.strptime(expected_last_time, '%Y/%m/%d %H:%M:%S')
            if table == 'BIDPEROFFER_D':
                cols = [dat_col, 'DUID', 'BIDTYPE']
                filter_cols = ('DUID', 'BIDTYPE')
                expected_number_of_columns = 3
            if table == 'BIDDAYOFFER_D':
                cols = [dat_col, 'DUID', 'BIDTYPE']
                filter_cols = ('DUID', 'BIDTYPE')
                expected_number_of_columns = 3
                expected_length = 2
                expected_last_time = '2015/05/01 00:00:00'
                expected_last_time = pd.Timestamp.strptime(expected_last_time, '%Y/%m/%d %H:%M:%S')
                expected_first_time = '2015/04/30 00:00:00'
                expected_first_time = pd.Timestamp.strptime(expected_first_time, '%Y/%m/%d %H:%M:%S')
            data = data_fetch_methods.dynamic_data_compiler(
                    start_time, end_time, table, defaults.raw_data_cache,
                    select_columns=cols,
                    filter_cols=filter_cols, filter_values=self.filter_values[table_type])
            data = data.reset_index(drop=True)
            print(table)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print('Passed')

    def test_dispatch_tables_end_of_month(self):
        start_time = '2013/07/31 21:00:00'
        end_time = '2013/08/01 00:00:00'
        for table in self.table_names:
            print('Testing {} returing values at end of month.'.format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            cols = [dat_col, self.table_types[table]]
            filter_cols = (self.table_types[table],)
            expected_length = 36
            expected_number_of_columns = 2
            expected_first_time = pd.Timestamp.strptime(start_time, '%Y/%m/%d %H:%M:%S') + timedelta(minutes=5)
            expected_last_time = pd.Timestamp.strptime(end_time, '%Y/%m/%d %H:%M:%S')
            if table in ['TRADINGLOAD', 'TRADINGPRICE', 'TRADINGREGIONSUM', 'TRADINGINTERCONNECT']:
                expected_length = 6
                expected_first_time = '2013/07/31 21:30:00'
                expected_first_time = pd.Timestamp.strptime(expected_first_time, '%Y/%m/%d %H:%M:%S')
            if table == 'BIDPEROFFER_D':
                cols = [dat_col, 'DUID', 'BIDTYPE']
                filter_cols = ('DUID', 'BIDTYPE')
                expected_number_of_columns = 3
            if table == 'BIDDAYOFFER_D':
                cols = [dat_col, 'DUID', 'BIDTYPE']
                filter_cols = ('DUID', 'BIDTYPE')
                expected_number_of_columns = 3
                expected_length = 1
                expected_last_time = expected_first_time.replace(hour=0, minute=0)
                expected_first_time = expected_first_time.replace(hour=0, minute=0)
            data = data_fetch_methods.dynamic_data_compiler(
                    start_time, end_time, table, defaults.raw_data_cache,
                    select_columns=cols,
                    filter_cols=filter_cols, filter_values=self.filter_values[table_type])
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print('Passed')

    def test_dispatch_tables_stradle_2_months(self):
        start_time = '2014/03/31 21:00:00'
        end_time = '2014/04/01 21:00:00'
        for table in self.table_names:
            print('Testing {} returing values from adjacent months.'.format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            cols = [dat_col, self.table_types[table]]
            filter_cols = (self.table_types[table],)
            expected_length = 288
            expected_number_of_columns = 2
            expected_first_time = pd.Timestamp.strptime(start_time, '%Y/%m/%d %H:%M:%S') + timedelta(minutes=5)
            expected_last_time = pd.Timestamp.strptime(end_time, '%Y/%m/%d %H:%M:%S')
            if table in ['TRADINGLOAD', 'TRADINGPRICE', 'TRADINGREGIONSUM', 'TRADINGINTERCONNECT']:
                expected_length = 48
                expected_first_time = '2014/03/31 21:30:00'
                expected_first_time = pd.Timestamp.strptime(expected_first_time, '%Y/%m/%d %H:%M:%S')
            if table == 'BIDPEROFFER_D':
                cols = [dat_col, 'DUID', 'BIDTYPE']
                filter_cols = ('DUID', 'BIDTYPE')
                expected_number_of_columns = 3
            if table == 'BIDDAYOFFER_D':
                cols = [dat_col, 'DUID', 'BIDTYPE']
                filter_cols = ('DUID', 'BIDTYPE')
                expected_number_of_columns = 3
                expected_length = 2
                expected_last_time = expected_last_time.replace(hour=0, minute=0)
                expected_first_time = expected_first_time.replace(hour=0, minute=0)
            data = data_fetch_methods.dynamic_data_compiler(
                    start_time, end_time, table, defaults.raw_data_cache,
                    select_columns=cols,
                    filter_cols=filter_cols, filter_values=self.filter_values[table_type])
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print('Passed')

    def test_dispatch_tables_start_of_year(self):
        start_time = '2014/01/01 00:00:00'
        end_time = '2014/01/01 01:00:00'
        for table in self.table_names:
            print('Testing {} returing values at start of year.'.format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            cols = [dat_col, self.table_types[table]]
            filter_cols = (self.table_types[table],)
            expected_length = 12
            expected_number_of_columns = 2
            expected_first_time = pd.Timestamp.strptime(start_time, '%Y/%m/%d %H:%M:%S') + timedelta(minutes=5)
            expected_last_time = pd.Timestamp.strptime(end_time, '%Y/%m/%d %H:%M:%S')
            if table in [ 'TRADINGLOAD', 'TRADINGPRICE', 'TRADINGREGIONSUM', 'TRADINGINTERCONNECT']:
                expected_length = 2
                expected_first_time = '2014/01/01 00:30:00'
                expected_first_time = pd.Timestamp.strptime(expected_first_time, '%Y/%m/%d %H:%M:%S')
            if table == 'BIDPEROFFER_D':
                cols = [dat_col, 'DUID', 'BIDTYPE']
                filter_cols = ('DUID', 'BIDTYPE')
                expected_number_of_columns = 3
            if table == 'BIDDAYOFFER_D':
                cols = [dat_col, 'DUID', 'BIDTYPE']
                filter_cols = ('DUID', 'BIDTYPE')
                expected_number_of_columns = 3
                expected_length = 1
                expected_last_time = expected_last_time.replace(hour=0, minute=0) - timedelta(days=1)
                expected_first_time = expected_first_time.replace(hour=0, minute=0) - timedelta(days=1)
            data = data_fetch_methods.dynamic_data_compiler(
                    start_time, end_time, table, defaults.raw_data_cache,
                    select_columns=cols,
                    filter_cols=filter_cols, filter_values=self.filter_values[table_type])
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print('Passed')

    def test_dispatch_tables_end_of_year(self):
        start_time = '2013/12/31 23:00:00'
        end_time = '2014/01/01 00:00:00'
        for table in self.table_names:
            print('Testing {} returing values at end of year.'.format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            cols = [dat_col, self.table_types[table]]
            filter_cols = (self.table_types[table],)
            expected_length = 12
            expected_number_of_columns = 2
            expected_first_time = pd.Timestamp.strptime(start_time, '%Y/%m/%d %H:%M:%S') + timedelta(minutes=5)
            expected_last_time = pd.Timestamp.strptime(end_time, '%Y/%m/%d %H:%M:%S')
            if table in [ 'TRADINGLOAD', 'TRADINGPRICE', 'TRADINGREGIONSUM', 'TRADINGINTERCONNECT']:
                expected_length = 2
                expected_first_time = '2013/12/31 23:30:00'
                expected_first_time = pd.Timestamp.strptime(expected_first_time, '%Y/%m/%d %H:%M:%S')
            if table == 'BIDPEROFFER_D':
                cols = [dat_col, 'DUID', 'BIDTYPE']
                filter_cols = ('DUID', 'BIDTYPE')
                expected_number_of_columns = 3
            if table == 'BIDDAYOFFER_D':
                cols = [dat_col, 'DUID', 'BIDTYPE']
                filter_cols = ('DUID', 'BIDTYPE')
                expected_number_of_columns = 3
                expected_length = 1
                expected_first_time = expected_first_time.replace(hour=0, minute=0)
                expected_last_time = expected_first_time
            data = data_fetch_methods.dynamic_data_compiler(
                    start_time, end_time, table, defaults.raw_data_cache,
                    select_columns=cols,
                    filter_cols=filter_cols, filter_values=self.filter_values[table_type])
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print('Passed')

    def test_dispatch_tables_stradle_years(self):
        start_time = '2017/12/31 23:00:00'
        end_time = '2018/01/01 01:00:00'
        for table in self.table_names:
            print('Testing {} returing values from adjacent years.'.format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            cols = [dat_col, self.table_types[table]]
            filter_cols = (self.table_types[table],)
            expected_length = 24
            expected_number_of_columns = 2
            expected_first_time = pd.Timestamp.strptime(start_time, '%Y/%m/%d %H:%M:%S') + timedelta(minutes=5)
            expected_last_time = pd.Timestamp.strptime(end_time, '%Y/%m/%d %H:%M:%S')
            if table in [ 'TRADINGLOAD', 'TRADINGPRICE', 'TRADINGREGIONSUM', 'TRADINGINTERCONNECT']:
                expected_length = 4
                expected_first_time = '2017/12/31 23:30:00'
                expected_first_time = pd.Timestamp.strptime(expected_first_time, '%Y/%m/%d %H:%M:%S')
            if table == 'BIDPEROFFER_D':
                cols = [dat_col, 'DUID', 'BIDTYPE']
                filter_cols = ('DUID', 'BIDTYPE')
                expected_number_of_columns = 3
            if table == 'BIDDAYOFFER_D':
                cols = [dat_col, 'DUID', 'BIDTYPE']
                filter_cols = ('DUID', 'BIDTYPE')
                expected_number_of_columns = 3
                expected_length = 1
                expected_first_time = expected_first_time.replace(hour=0, minute=0)
                expected_last_time = expected_first_time
            data = data_fetch_methods.dynamic_data_compiler(
                    start_time, end_time, table, defaults.raw_data_cache,
                    select_columns=cols,
                    filter_cols=filter_cols, filter_values=self.filter_values[table_type])
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print('Passed')


class TestDynamicDataCompilerWithEffectiveDateFiltering(unittest.TestCase):
    def setUp(self):
        self.table_names = ['GENCONDATA', 'SPDREGIONCONSTRAINT', 'SPDCONNECTIONPOINTCONSTRAINT',
                            'SPDINTERCONNECTORCONSTRAINT']

    def test_filtering_for_one_interval_returns(self):
        start_time = '2017/05/20 23:00:00'
        end_time = '2017/05/20 23:05:00'
        for table in self.table_names:
            print('Testing {} returing values for 1 interval.'.format(table))
            data = data_fetch_methods.dynamic_data_compiler(
                    start_time, end_time, table, defaults.raw_data_cache,
                    select_columns=defaults.table_primary_keys[table])
            group_cols = [col for col in defaults.table_primary_keys[table] if col != 'EFFECTIVEDATE']
            contains_duplicates = data.duplicated(group_cols).any()
            self.assertEqual(False, contains_duplicates)
            not_empty = data.shape[0] > 0
            self.assertEqual(True, not_empty)
            print('Passed')


class TestDynamicDataCompilerWithStartDateFiltering(unittest.TestCase):
    def setUp(self):
        self.table_names = ['DUDETAILSUMMARY']

    def test_filtering_for_one_interval_returns(self):
        start_time = '2017/05/20 23:00:00'
        end_time = '2017/05/20 23:05:00'
        for table in self.table_names:
            print('Testing {} returing values for 1 interval.'.format(table))
            data = data_fetch_methods.dynamic_data_compiler(
                    start_time, end_time, table, defaults.raw_data_cache,
                    select_columns=defaults.table_primary_keys[table] + ['END_DATE'])
            group_cols = [col for col in defaults.table_primary_keys[table] if col != 'START_DATE']
            contains_duplicates = data.duplicated(group_cols).any()
            self.assertEqual(False, contains_duplicates)
            not_empty = data.shape[0] > 0
            self.assertEqual(True, not_empty)
            print('Passed')


class TestFACS4SecondData(unittest.TestCase):
    def setUp(self):
        pass

    def test_dispatch_tables_start_of_month(self):
        table = 'FCAS_4_SECOND'
        start_time = '2015/05/01 00:00:00'
        end_time = '2015/05/01 00:05:08'
        print('Testing {} returing values at start of month.'.format(table))
        dat_col = defaults.primary_date_columns[table]
        cols = [dat_col, 'ELEMENTNUMBER', 'VARIABLENUMBER']
        filter_cols = ('ELEMENTNUMBER', 'VARIABLENUMBER')
        expected_length = 77
        expected_number_of_columns = 3
        expected_firt_time = pd.Timestamp.strptime(start_time, '%Y/%m/%d %H:%M:%S') + timedelta(seconds=4)
        expected_last_time = pd.Timestamp.strptime(end_time, '%Y/%m/%d %H:%M:%S')
        data = data_fetch_methods.dynamic_data_compiler(
                start_time, end_time, table, defaults.raw_data_cache,
                select_columns=cols,
                filter_cols=filter_cols, filter_values=(['1'], ['3']))
        data = data.sort_values(dat_col)
        data = data.reset_index(drop=True)
        self.assertEqual(expected_length, data.shape[0])
        self.assertEqual(expected_number_of_columns, data.shape[1])
        self.assertEqual(expected_firt_time, data[dat_col][0])
        self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
        print('Passed')

    def test_fcas_tables_end_of_month(self):
        table = 'FCAS_4_SECOND'
        start_time = '2013/07/31 23:55:03'
        end_time = '2013/08/01 00:00:04'
        print('Testing {} returing values at end of month.'.format(table))
        dat_col = defaults.primary_date_columns[table]
        cols = [dat_col, 'ELEMENTNUMBER', 'VARIABLENUMBER']
        filter_cols = ('ELEMENTNUMBER', 'VARIABLENUMBER')
        expected_length = 75
        expected_number_of_columns = 3
        expected_firt_time = pd.Timestamp.strptime(start_time, '%Y/%m/%d %H:%M:%S') + timedelta(seconds=4)
        expected_last_time = pd.Timestamp.strptime(end_time, '%Y/%m/%d %H:%M:%S')
        data = data_fetch_methods.dynamic_data_compiler(
            start_time, end_time, table, defaults.raw_data_cache,
            select_columns=cols,
            filter_cols=filter_cols, filter_values=(['1'], ['3']))
        data = data.sort_values(dat_col)
        data = data.reset_index(drop=True)
        self.assertEqual(expected_length, data.shape[0])
        self.assertEqual(expected_number_of_columns, data.shape[1])
        self.assertEqual(expected_firt_time, data[dat_col][0])
        self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
        print('Passed')

    def test_fcas_tables_stradle_2_months(self):
        table = 'FCAS_4_SECOND'
        start_time = '2013/07/31 23:55:03'
        end_time = '2013/08/01 00:05:04'
        print('Testing {} returing values at end of month.'.format(table))
        dat_col = defaults.primary_date_columns[table]
        cols = [dat_col, 'ELEMENTNUMBER', 'VARIABLENUMBER']
        filter_cols = ('ELEMENTNUMBER', 'VARIABLENUMBER')
        expected_length = 150
        expected_number_of_columns = 3
        expected_firt_time = pd.Timestamp.strptime(start_time, '%Y/%m/%d %H:%M:%S') + timedelta(seconds=4)
        expected_last_time = pd.Timestamp.strptime(end_time, '%Y/%m/%d %H:%M:%S')
        data = data_fetch_methods.dynamic_data_compiler(
            start_time, end_time, table, defaults.raw_data_cache,
            select_columns=cols,
            filter_cols=filter_cols, filter_values=(['1'], ['3']))
        data = data.sort_values(dat_col)
        data = data.reset_index(drop=True)
        self.assertEqual(expected_length, data.shape[0])
        self.assertEqual(expected_number_of_columns, data.shape[1])
        self.assertEqual(expected_firt_time, data[dat_col][0])
        self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
        print('Passed')

    def test_fcas_tables_end_of_year(self):
        table = 'FCAS_4_SECOND'
        start_time = '2013/12/31 23:50:04'
        end_time = '2014/01/01 00:00:00'
        print('Testing {} returing values at end of year.'.format(table))
        dat_col = defaults.primary_date_columns[table]
        cols = [dat_col, 'ELEMENTNUMBER', 'VARIABLENUMBER']
        filter_cols = ('ELEMENTNUMBER', 'VARIABLENUMBER')
        expected_length = 149
        expected_number_of_columns = 3
        expected_firt_time = pd.Timestamp.strptime(start_time, '%Y/%m/%d %H:%M:%S') + timedelta(seconds=4)
        expected_last_time = pd.Timestamp.strptime(end_time, '%Y/%m/%d %H:%M:%S')
        data = data_fetch_methods.dynamic_data_compiler(
            start_time, end_time, table, defaults.raw_data_cache,
            select_columns=cols,
            filter_cols=filter_cols, filter_values=(['1'], ['3']))
        data = data.sort_values(dat_col)
        data = data.reset_index(drop=True)
        self.assertEqual(expected_length, data.shape[0])
        self.assertEqual(expected_number_of_columns, data.shape[1])
        self.assertEqual(expected_firt_time, data[dat_col][0])
        self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
        print('Passed')

    def test_dispatch_tables_stradle_years(self):
        table = 'FCAS_4_SECOND'
        start_time = '2011/12/31 23:55:04'
        end_time = '2012/01/01 00:05:00'
        print('Testing {} returing values from adjacent years.'.format(table))
        dat_col = defaults.primary_date_columns[table]
        cols = [dat_col, 'ELEMENTNUMBER', 'VARIABLENUMBER']
        filter_cols = ('ELEMENTNUMBER', 'VARIABLENUMBER')
        expected_length = 149
        expected_number_of_columns = 3
        expected_firt_time = pd.Timestamp.strptime(start_time, '%Y/%m/%d %H:%M:%S') + timedelta(seconds=4)
        expected_last_time = pd.Timestamp.strptime(end_time, '%Y/%m/%d %H:%M:%S')
        data = data_fetch_methods.dynamic_data_compiler(
            start_time, end_time, table, defaults.raw_data_cache,
            select_columns=cols,
            filter_cols=filter_cols, filter_values=(['1'], ['3']))
        data = data.sort_values(dat_col)
        data = data.reset_index(drop=True)
        self.assertEqual(expected_length, data.shape[0])
        self.assertEqual(expected_number_of_columns, data.shape[1])
        self.assertEqual(expected_firt_time, data[dat_col][0])
        self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
        print('Passed')


class TestStaticTables(unittest.TestCase):
    def setUp(self):
        pass

    def test_fcas_elements_table(self):
        start_time = ''
        end_time = ''
        table = 'ELEMENTS_FCAS_4_SECOND'
        cols = ['ELEMENTNUMBER', 'EMSNAME']
        filter_cols = ('ELEMENTNUMBER',)
        data = data_fetch_methods.static_table( start_time, end_time, table, defaults.raw_data_cache,
            select_columns=cols, filter_cols=filter_cols, filter_values=(['1'],))
        expected_length = 1
        expected_number_of_columns = 2
        self.assertEqual(expected_length, data.shape[0])
        self.assertEqual(expected_number_of_columns, data.shape[1])
        print('Passed')

    def test_fcas_variable_table(self):
        start_time = ''
        end_time = ''
        table = 'VARIABLES_FCAS_4_SECOND'
        cols = ['VARIABLENUMBER', 'VARIABLETYPE']
        filter_cols = ('VARIABLENUMBER',)
        data = data_fetch_methods.static_table( start_time, end_time, table, defaults.raw_data_cache,
            select_columns=cols, filter_cols=filter_cols, filter_values=(['2'],))
        expected_length = 1
        expected_number_of_columns = 2
        self.assertEqual(expected_length, data.shape[0])
        self.assertEqual(expected_number_of_columns, data.shape[1])
        print('Passed')

    def test_registration_list(self):
        start_time = ''
        end_time = ''
        table = 'Generators and Scheduled Loads'
        cols = ['DUID', 'Technology Type - Primary']
        filter_cols = ('DUID',)
        data = data_fetch_methods.static_table_xl( start_time, end_time, table, defaults.raw_data_cache,
            select_columns=cols, filter_cols=filter_cols, filter_values=(['AGLHAL'],))
        expected_length = 1
        expected_number_of_columns = 2
        self.assertEqual(expected_length, data.shape[0])
        self.assertEqual(expected_number_of_columns, data.shape[1])
        print('Passed')


class TestCustomTables(unittest.TestCase):
    def setUp(self):
        pass

    def test_dispatch_tables_stradle_years(self):
        table = 'FCAS_4s_SCADA_MAP'
        start_time = '2012/01/01 00:05:00'
        end_time = '2012/01/01 01:00:00'
        print('Testing custom table {}.'.format(table))
        data = custom_tables.fcas4s_scada_match(start_time, end_time, table, defaults.raw_data_cache)
        data = data.reset_index(drop=True)
        contains_duplicates = data.duplicated(['MARKETNAME']).any()
        self.assertEqual(False, contains_duplicates)
        contains_duplicates = data.duplicated(['ELEMENTNUMBER']).any()
        self.assertEqual(False, contains_duplicates)
        not_empty = data.shape[0] > 0
        self.assertEqual(True, not_empty)
        print('Passed')



import requests
import zipfile
import io
from nemosis import defaults



def run(year, month, day, index, filename, down_load_to):
    """This function"""

    url = defaults.aemo_data_url
    # Add the year and month information to the generic AEMO data url
    url_formatted = format_aemo_url(url, year, month, filename)

    # Perform the download, unzipping saving of the file
    try:
        download_unzip_csv(url_formatted, down_load_to, filename)
    except:
        print('Warning {} not downloaded'.format(filename))


def run_fcas4s(year, month, day, index, filename, down_load_to):
    """This function"""

    # Add the year and month information to the generic AEMO data url
    url_formatted_latest = defaults.fcas_4_url.format(year, month, day, index)
    url_formatted_hist = defaults.fcas_4_url_hist.format(year, year, month, year, month, day, index)


    # Perform the download, unzipping saving of the file
    try:
        download_unzip_csv(url_formatted_latest, down_load_to, filename)
    except:
        try:
            download_unzip_csv(url_formatted_hist, down_load_to, filename)
        except:
            print('Warning {} not downloaded'.format(filename))


def download_unzip_csv(url, down_load_to, filename):
    """This function downloads a zipped csv using a url, extracts the csv and saves it a specified location and with
    a specified filename"""
    r = requests.get(url)
    z = zipfile.ZipFile(io.BytesIO(r.content))
    z.extractall(down_load_to)


def download_csv(url, down_load_to, path_and_name):
    """This function downloads a zipped csv using a url, extracts the csv and saves it a specified location and with
    a specified filename"""
    r = requests.get(url)
    with open(path_and_name, 'wb') as f:
        f.write(r.content)


def download_xl(url, down_load_to, path_and_name):
    """This function downloads a zipped csv using a url, extracts the csv and saves it a specified location and with
    a specified filename"""
    r = requests.get(url)
    with open(path_and_name, 'wb') as f:
        f.write(r.content)

def format_aemo_url(url, year, month, filename):
    """This fills in the missing information in the AEMO url so data for the right month, year and file name are
    downloaded"""
    year = str(year)
    return url.format(year, year, month, filename[:-4])




import unittest
import pandas as pd
from nemosis import filters
from pandas.util.testing import assert_frame_equal
from datetime import datetime
import numpy as np


class TestFiltersStartDate(unittest.TestCase):
    def setUp(self):
        self.start_date_data = \
            pd.DataFrame({'START_DATE': ['2011/01/01 00:00:00', '2015/01/01 00:00:00'],
                          'END_DATE': ['2015/01/01 00:00:00', '2015/07/01 00:12:00']})

    def test_start_date_pick_first_of_two(self):
        start_time = datetime.strptime('2014/06/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2014/09/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_start_and_end_date(self.start_date_data, start_time=start_time, end_time=end_time)
        aim = pd.DataFrame({'START_DATE': ['2011/01/01 00:00:00'], 'END_DATE': ['2015/01/01 00:00:00']})
        aim['START_DATE'] = pd.to_datetime(aim['START_DATE'], format='%Y/%m/%d %H:%M:%S')
        aim['END_DATE'] = pd.to_datetime(aim['END_DATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_start_date_pick_second_of_two(self):
        start_time = datetime.strptime('2015/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2015/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_start_and_end_date(self.start_date_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'START_DATE': ['2015/01/01 00:00:00'], 'END_DATE': ['2015/07/01 00:12:00']})
        aim['START_DATE'] = pd.to_datetime(aim['START_DATE'], format='%Y/%m/%d %H:%M:%S')
        aim['END_DATE'] = pd.to_datetime(aim['END_DATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_start_date_pick_two_of_two_by_overlaping_interval(self):
        start_time = datetime.strptime('2011/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2015/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_start_and_end_date(self.start_date_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'START_DATE': ['2011/01/01 00:00:00', '2015/01/01 00:00:00'],
                          'END_DATE': ['2015/01/01 00:00:00', '2015/07/01 00:12:00']})
        aim['START_DATE'] = pd.to_datetime(aim['START_DATE'], format='%Y/%m/%d %H:%M:%S')
        aim['END_DATE'] = pd.to_datetime(aim['END_DATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_start_date_pick_none_of_two_by_overshooting_date(self):
        start_time = datetime.strptime('2018/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2019/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_start_and_end_date(self.start_date_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'START_DATE': [],
                          'END_DATE': []})
        aim['START_DATE'] = pd.to_datetime(aim['START_DATE'], format='%Y/%m/%d %H:%M:%S')
        aim['END_DATE'] = pd.to_datetime(aim['END_DATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_start_date_pick_none_of_two_by_undershooting_date(self):
        start_time = datetime.strptime('2010/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2010/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_start_and_end_date(self.start_date_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'START_DATE': [],
                          'END_DATE': []})
        aim['START_DATE'] = pd.to_datetime(aim['START_DATE'], format='%Y/%m/%d %H:%M:%S')
        aim['END_DATE'] = pd.to_datetime(aim['END_DATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)


class TestFiltersEffectiveDate(unittest.TestCase):
    def setUp(self):
        self.last_changed_data = pd.DataFrame({'EFFECTIVEDATE': ['2011/01/01 00:00:00', '2015/01/01 00:00:00']})

    def test_start_date_pick_first_of_two(self):
        start_time = datetime.strptime('2014/06/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2014/09/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_effective_date(self.last_changed_data, start_time=start_time, end_time=end_time)
        aim = pd.DataFrame({'EFFECTIVEDATE': ['2011/01/01 00:00:00']})
        aim['EFFECTIVEDATE'] = pd.to_datetime(aim['EFFECTIVEDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_start_date_two_of_two_with_date_window_on_second(self):
        start_time = datetime.strptime('2015/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2015/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_effective_date(self.last_changed_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'EFFECTIVEDATE': ['2011/01/01 00:00:00', '2015/01/01 00:00:00']})
        aim['EFFECTIVEDATE'] = pd.to_datetime(aim['EFFECTIVEDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_start_date_pick_two_of_two_by_overlaping_interval(self):
        start_time = datetime.strptime('2011/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2015/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_effective_date(self.last_changed_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'EFFECTIVEDATE': ['2011/01/01 00:00:00', '2015/01/01 00:00:00']})
        aim['EFFECTIVEDATE'] = pd.to_datetime(aim['EFFECTIVEDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_start_date_pick_two_of_two_by_overshooting_date(self):
        start_time = datetime.strptime('2018/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2019/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_effective_date(self.last_changed_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'EFFECTIVEDATE': ['2011/01/01 00:00:00', '2015/01/01 00:00:00']})
        aim['EFFECTIVEDATE'] = pd.to_datetime(aim['EFFECTIVEDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_start_date_pick_none_of_two_by_undershooting_date(self):
        start_time = datetime.strptime('2010/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2010/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_effective_date(self.last_changed_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'EFFECTIVEDATE': []})
        aim['EFFECTIVEDATE'] = pd.to_datetime(aim['EFFECTIVEDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)


class TestFiltersSettlementDate(unittest.TestCase):
    def setUp(self):
        self.settlement_date_data = \
            pd.DataFrame({'SETTLEMENTDATE': ['2011/01/01 00:00:00', '2015/01/01 00:00:00']})

    def test_settlement_date_pick_first_of_two(self):
        start_time = datetime.strptime('2010/06/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2011/09/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_settlementdate(self.settlement_date_data, start_time=start_time, end_time=end_time)
        aim = pd.DataFrame({'SETTLEMENTDATE': ['2011/01/01 00:00:00']})
        aim['SETTLEMENTDATE'] = pd.to_datetime(aim['SETTLEMENTDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_settlement_date_pick_second_of_two(self):
        start_time = datetime.strptime('2014/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2015/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_settlementdate(self.settlement_date_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'SETTLEMENTDATE': ['2015/01/01 00:00:00']})
        aim['SETTLEMENTDATE'] = pd.to_datetime(aim['SETTLEMENTDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_settlement_date_pick_two_of_two_by_overlaping_interval(self):
        start_time = datetime.strptime('2010/12/31 23:59:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2015/05/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        aim = pd.DataFrame({'SETTLEMENTDATE': ['2011/01/01 00:00:00', '2015/01/01 00:00:00']})
        result = filters.filter_on_settlementdate(self.settlement_date_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim['SETTLEMENTDATE'] = pd.to_datetime(aim['SETTLEMENTDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_settlement_date_pick_none_of_two_by_overshooting_date(self):
        start_time = datetime.strptime('2018/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2019/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_settlementdate(self.settlement_date_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'SETTLEMENTDATE': []})
        aim['SETTLEMENTDATE'] = pd.to_datetime(aim['SETTLEMENTDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_settlement_date_pick_none_of_two_by_undershooting_date(self):
        start_time = datetime.strptime('2010/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2010/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_settlementdate(self.settlement_date_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'SETTLEMENTDATE': []})
        aim['SETTLEMENTDATE'] = pd.to_datetime(aim['SETTLEMENTDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_settlement_date_end_date_exclusive_by_undershooting(self):
        start_time = datetime.strptime('2010/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2010/12/31 23:59:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_settlementdate(self.settlement_date_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'SETTLEMENTDATE': []})
        aim['SETTLEMENTDATE'] = pd.to_datetime(aim['SETTLEMENTDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)


class TestFiltersTimeStamp(unittest.TestCase):
    def setUp(self):
        self.time_stamp_data = \
            pd.DataFrame({'TIMESTAMP': ['2011/01/01 00:00:00', '2015/01/01 00:00:00']})

    def test_time_stamp_pick_first_of_two(self):
        start_time = datetime.strptime('2010/06/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2011/09/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_timestamp(self.time_stamp_data, start_time=start_time, end_time=end_time)
        aim = pd.DataFrame({'TIMESTAMP': ['2011/01/01 00:00:00']})
        aim['TIMESTAMP'] = pd.to_datetime(aim['TIMESTAMP'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_time_stamp_pick_second_of_two(self):
        start_time = datetime.strptime('2014/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2015/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_timestamp(self.time_stamp_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'TIMESTAMP': ['2015/01/01 00:00:00']})
        aim['TIMESTAMP'] = pd.to_datetime(aim['TIMESTAMP'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_time_stamp_pick_two_of_two_by_overlaping_interval(self):
        start_time = datetime.strptime('2010/12/31 23:59:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2015/05/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        aim = pd.DataFrame({'TIMESTAMP': ['2011/01/01 00:00:00', '2015/01/01 00:00:00']})
        result = filters.filter_on_timestamp(self.time_stamp_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim['TIMESTAMP'] = pd.to_datetime(aim['TIMESTAMP'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_time_stamp_pick_none_of_two_by_overshooting_date(self):
        start_time = datetime.strptime('2018/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2019/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_timestamp(self.time_stamp_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'TIMESTAMP': []})
        aim['TIMESTAMP'] = pd.to_datetime(aim['TIMESTAMP'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_time_stamp_pick_none_of_two_by_undershooting_date(self):
        start_time = datetime.strptime('2010/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2010/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_timestamp(self.time_stamp_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'TIMESTAMP': []})
        aim['TIMESTAMP'] = pd.to_datetime(aim['TIMESTAMP'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_time_stamp_end_date_exclusive_by_undershooting(self):
        start_time = datetime.strptime('2010/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2010/12/31 23:59:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_timestamp(self.time_stamp_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'TIMESTAMP': []})
        aim['TIMESTAMP'] = pd.to_datetime(aim['TIMESTAMP'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)


class TestFiltersIntervalDatetime(unittest.TestCase):
    def setUp(self):
        self.interval_datetime_data = \
            pd.DataFrame({'INTERVAL_DATETIME': ['2011/01/01 00:00:00', '2015/01/01 00:00:00']})

    def test_interval_datetime_pick_first_of_two(self):
        start_time = datetime.strptime('2010/06/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2011/09/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_interval_datetime(self.interval_datetime_data, start_time=start_time, end_time=end_time)
        aim = pd.DataFrame({'INTERVAL_DATETIME': ['2011/01/01 00:00:00']})
        aim['INTERVAL_DATETIME'] = pd.to_datetime(aim['INTERVAL_DATETIME'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_interval_datetime_pick_second_of_two(self):
        start_time = datetime.strptime('2014/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2015/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_interval_datetime(self.interval_datetime_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'INTERVAL_DATETIME': ['2015/01/01 00:00:00']})
        aim['INTERVAL_DATETIME'] = pd.to_datetime(aim['INTERVAL_DATETIME'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_interval_datetime_pick_two_of_two_by_overlaping_interval(self):
        start_time = datetime.strptime('2010/12/31 23:59:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2015/05/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        aim = pd.DataFrame({'INTERVAL_DATETIME': ['2011/01/01 00:00:00', '2015/01/01 00:00:00']})
        result = filters.filter_on_interval_datetime(self.interval_datetime_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim['INTERVAL_DATETIME'] = pd.to_datetime(aim['INTERVAL_DATETIME'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_interval_datetime_pick_none_of_two_by_overshooting_date(self):
        start_time = datetime.strptime('2018/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2019/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_interval_datetime(self.interval_datetime_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'INTERVAL_DATETIME': []})
        aim['INTERVAL_DATETIME'] = pd.to_datetime(aim['INTERVAL_DATETIME'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_interval_datetime_stamp_pick_none_of_two_by_undershooting_date(self):
        start_time = datetime.strptime('2010/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2010/05/01 00:13:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_interval_datetime(self.interval_datetime_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'INTERVAL_DATETIME': []})
        aim['INTERVAL_DATETIME'] = pd.to_datetime(aim['INTERVAL_DATETIME'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_interval_datetime_end_date_exclusive_by_undershooting(self):
        start_time = datetime.strptime('2010/04/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2010/12/31 23:59:00', '%Y/%m/%d %H:%M:%S')
        result = filters.filter_on_interval_datetime(self.interval_datetime_data, start_time=start_time, end_time=end_time)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'INTERVAL_DATETIME': []})
        aim['INTERVAL_DATETIME'] = pd.to_datetime(aim['INTERVAL_DATETIME'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)


class TestFiltersColumnValue(unittest.TestCase):
    def setUp(self):
        self.int_filter_data = pd.DataFrame({'INTCOL': [1, 10, 100, -5, 0, 10, 10]})
        self.int_and_string_filter_data = pd.DataFrame({'INTCOL': [1, 10, 100, -5, 0, 10, 10],
                                                        'STRINGCOL': ['1', '10', '100', '-5', '0', '10', '10']})

    def test_filter_one_col_one_value_positive_ints(self):
        filter_cols = ('INTCOL', )
        filter_values = ([10],)
        result = \
            filters.filter_on_column_value(self.int_filter_data, filter_cols, filter_values).reset_index(drop=True)
        aim = pd.DataFrame({'INTCOL': [10, 10, 10]})
        assert_frame_equal(aim, result)

    def test_filter_one_col_two_values_positive_and_negative_ints(self):
        filter_cols = ('INTCOL', )
        filter_values = ([10, -5],)
        result = \
            filters.filter_on_column_value(self.int_filter_data, filter_cols, filter_values).reset_index(drop=True)
        aim = pd.DataFrame({'INTCOL': [10, -5, 10, 10]})
        assert_frame_equal(aim, result)

    def test_filter_two_cols_one_value_each_not_matching(self):
        filter_cols = ('INTCOL', 'STRINGCOL' )
        filter_values = ([10], ['100'])
        result = \
            filters.filter_on_column_value(self.int_and_string_filter_data, filter_cols, filter_values).reset_index(drop=True)
        aim = pd.DataFrame({'INTCOL': [], 'STRINGCOL': []})
        aim = aim.astype(dtype={'INTCOL': np.int64, 'STRINGCOL': str})
        assert_frame_equal(aim, result)

    def test_filter_two_cols_one_value_each_matching(self):
        filter_cols = ('INTCOL', 'STRINGCOL' )
        filter_values = ([10], ['10'])
        result = \
            filters.filter_on_column_value(self.int_and_string_filter_data, filter_cols, filter_values).reset_index(drop=True)
        aim = pd.DataFrame({'INTCOL': [10, 10, 10], 'STRINGCOL': ['10', '10', '10']})
        assert_frame_equal(aim, result)

    def test_filter_just_one_of_two_cols(self):
        filter_cols = ('INTCOL', )
        filter_values = ([10],)
        result = \
            filters.filter_on_column_value(self.int_and_string_filter_data, filter_cols, filter_values).reset_index(drop=True)
        aim = pd.DataFrame({'INTCOL': [10, 10, 10], 'STRINGCOL': ['10', '10', '10']})
        assert_frame_equal(aim, result)

    def test_filter_one_empty_values_returns_empty_data_frame(self):
        filter_cols = ('INTCOL', )
        filter_values = ([],)
        result = \
            filters.filter_on_column_value(self.int_and_string_filter_data, filter_cols, filter_values).reset_index(drop=True)
        aim = pd.DataFrame({'INTCOL': [], 'STRINGCOL': []})
        aim = aim.astype(dtype={'INTCOL': np.int64, 'STRINGCOL': str})
        assert_frame_equal(aim, result)


if __name__ == '__main__':
    unittest.main()






import pandas as pd
from datetime import datetime, timedelta
from nemosis import defaults



def dispatch_date_setup(start_time, end_time):
    start_time = datetime.strptime(start_time, '%Y/%m/%d %H:%M:%S')
    start_time = start_time - timedelta(hours=4)
    start_time = start_time.replace(hour=0, minute=0)
    start_time = start_time - timedelta(seconds=1)
    start_time = datetime.isoformat(start_time).replace('-', '/').replace('T', ' ')
    end_time = datetime.strptime(end_time, '%Y/%m/%d %H:%M:%S')
    end_time = end_time - timedelta(hours=4, seconds=1)
    end_time = datetime.isoformat(end_time).replace('-', '/').replace('T', ' ')
    end_time = end_time[:10]
    date_padding = ' 00:00:00'
    end_time = end_time + date_padding
    return start_time, end_time


def dispatch_half_hour_setup(start_time, end_time):
    start_time = datetime.strptime(start_time, '%Y/%m/%d %H:%M:%S')
    start_time = datetime(year=start_time.year, month=start_time.month, day=start_time.day, hour=start_time.hour,
                          minute=((start_time.minute // 30) * 30))
    start_time = start_time.isoformat().replace('T', ' ').replace('-', '/')
    return start_time, end_time


def fcas4s_finalise(data, start_time, table_name):
    for column in data.select_dtypes(['object']).columns:
        data[column] = data[column].map(lambda x: x.strip())
    return data


def most_recent_records_before_start_time(data, start_time, table_name):
    date_col = defaults.primary_date_columns[table_name]
    group_cols = defaults.effective_date_group_col[table_name]
    records_from_after_start = data[data[date_col] >= start_time].copy()
    records_from_before_start = data[data[date_col] < start_time].copy()
    records_from_before_start = records_from_before_start.sort_values(date_col)
    if len(group_cols) > 0:
        most_recent_from_before_start = records_from_before_start.groupby(group_cols, as_index=False).last()
        group_cols = group_cols + [date_col]
        most_recent_from_before_start = pd.merge(most_recent_from_before_start.loc[:, group_cols],
                                                 records_from_before_start, 'inner', group_cols)
    else:
        most_recent_from_before_start = records_from_before_start.tail(1)

    mod_table = pd.concat([records_from_after_start, most_recent_from_before_start], sort=False)
    return mod_table


def drop_duplicates_by_primary_key(data, start_time, table_name):
    data = data.drop_duplicates(defaults.table_primary_keys[table_name])
    return data


import unittest
from nemosis import processing_info_maps
from nemosis import data_fetch_methods
from nemosis import defaults
import pandas as pd
from datetime import datetime
from nemosis import query_wrapers


class TestSearchTypeValidity(unittest.TestCase):
    def setUp(self):
        pass

    def test_start_to_end_no_duplication_between_batches(self):
        for table_name in processing_info_maps.search_type.keys():
            if processing_info_maps.search_type[table_name] == 'start_to_end':
                print('Validating start_to_end type for table {}'.format(table_name))
                start_test_window = defaults.nem_data_model_start_time
                start_test_window = '2018/01/01 00:00:00'
                start_time = datetime.strptime(start_test_window, '%Y/%m/%d %H:%M:%S')
                end_time = datetime.strptime('2018/05/01 00:00:00', '%Y/%m/%d %H:%M:%S')
                if table_name == 'FCAS_4_SECOND':
                    start_test_window = '2015/01/01 00:00:00'
                    end_time = datetime.strptime('2015/01/02 00:00:00', '%Y/%m/%d %H:%M:%S')
                start_search = datetime.strptime(start_test_window, '%Y/%m/%d %H:%M:%S')
                data_tables = data_fetch_methods.dynamic_data_fetch_loop(
                    start_search=start_search, start_time=start_time,
                    end_time=end_time, table_name=table_name, raw_data_location='E:/raw_aemo_data',
                    select_columns=defaults.table_primary_keys[table_name], date_filter=None,
                    search_type='start_to_end')
                all_data = pd.concat(data_tables, sort=False)
                contains_duplicates = all_data.duplicated().any()
                self.assertEqual(False, contains_duplicates, 'table {}'.format(table_name))
                print('Type valid, no duplicates found.')

    def test_start_to_end_has_settlement_or_interval_col(self):
        for table_name in processing_info_maps.search_type.keys():
            if processing_info_maps.search_type[table_name] == 'start_to_end':
                has_settlement_date_col = 'SETTLEMENTDATE' in defaults.table_columns[table_name]
                has_interval_datetime_col = 'INTERVAL_DATETIME' in defaults.table_columns[table_name]
                has_interval_timestamp_col = 'TIMESTAMP' in defaults.table_columns[table_name]
                has_either = has_interval_datetime_col or has_settlement_date_col or has_interval_timestamp_col
                self.assertEqual(True, has_either)
                print('{} is valid candidate for type start_to_end as there is a SETTLEMENTDATE, ' 
                      'INTERVAL_DATETIME or TIMESTAMP column to filter on.'
                      .format(table_name))

    def test_all_no_duplication_between_batches(self):
        for table_name in processing_info_maps.search_type.keys():
            if processing_info_maps.search_type[table_name] == 'all':
                print('Validating all type for table {}'.format(table_name))
                if table_name in ['GENCONDATA', 'SPDCONNECTIONPOINTCONSTRAINT', 'SPDINTERCONNECTORCONSTRAINT',
                                  'DUDETAILSUMMARY', 'LOSSMODEL', 'LOSSFACTORMODEL', 'MNSP_DAYOFFER',
                                  'MNSP_PEROFFER', 'MNSP_INTERCONNECTOR', 'INTERCONNECTOR', 'INTERCONNECTORCONSTRAINT',
                                  'DUDETAIL']:
                    print('{} is known to contain duplicate entries and is exempted from this test, a finalise '
                          'data processing step is included in dynamic data fetch to clean up these duplicates.'
                          .format(table_name))
                    continue
                start_test_window = defaults.nem_data_model_start_time
                #start_test_window = '2018/01/01 00:00:00'
                start_time = datetime.strptime(start_test_window, '%Y/%m/%d %H:%M:%S')
                end_time = datetime.strptime('2018/01/01 00:00:00', '%Y/%m/%d %H:%M:%S')
                start_search = datetime.strptime(start_test_window, '%Y/%m/%d %H:%M:%S')
                data_tables = data_fetch_methods.dynamic_data_fetch_loop(
                    start_search=start_search, start_time=start_time,
                    end_time=end_time, table_name=table_name, raw_data_location='E:/raw_aemo_data',
                    select_columns=defaults.table_primary_keys[table_name], date_filter=None,
                    search_type='all')
                all_data = pd.concat(data_tables, sort=False)
                contains_duplicates = all_data.duplicated().any()
                self.assertEqual(False, contains_duplicates, 'table {}'.format(table_name))
                print('Type valid, no duplicates found.')

    def test_all_no_duplication_between_batches_with_finalise_step(self):
        for table_name in processing_info_maps.search_type.keys():
            if processing_info_maps.search_type[table_name] == 'all':
                print('Testing duplicate removal for table {}'.format(table_name))
                start_test_window = defaults.nem_data_model_start_time
                #start_test_window = '2018/01/01 00:00:00'
                start_time = datetime.strptime(start_test_window, '%Y/%m/%d %H:%M:%S')
                end_time = datetime.strptime('2018/01/01 00:00:00', '%Y/%m/%d %H:%M:%S')
                start_search = datetime.strptime(start_test_window, '%Y/%m/%d %H:%M:%S')
                data_tables = data_fetch_methods.dynamic_data_fetch_loop(
                    start_search=start_search, start_time=start_time,
                    end_time=end_time, table_name=table_name, raw_data_location='E:/raw_aemo_data',
                    select_columns=defaults.table_primary_keys[table_name], date_filter=None,
                    search_type='all')
                all_data = pd.concat(data_tables, sort=False)
                all_data = query_wrapers.drop_duplicates_by_primary_key(all_data, start_time, table_name)
                contains_duplicates = all_data.duplicated().any()
                self.assertEqual(False, contains_duplicates)
                print('Type valid, no duplicates found.')

    def test_start_to_end_has_no_settlement_interval_or_timestamp_col(self):
        for table_name in processing_info_maps.search_type.keys():
            if processing_info_maps.search_type[table_name] == 'all':
                has_settlement_date_col = 'SETTLEMENTDATE' in defaults.table_columns[table_name]
                has_interval_datetime_col = 'INTERVAL_DATETIME' in defaults.table_columns[table_name]
                has_interval_timestamp_col = 'TIMESTAMP' in defaults.table_columns[table_name]
                has_either = has_interval_datetime_col or has_settlement_date_col or has_interval_timestamp_col
                self.assertEqual(False, has_either, 'table {}'.format(table_name))
                print('{} is valid candidate for type all as there is not a SETTLEMENTDATE, ' 
                      'INTERVAL_DATETIME or TIMESTAMP column to filter on'
                      .format(table_name))

    def test_last_contains_data_from_first(self):
        for table_name in processing_info_maps.search_type.keys():
            if processing_info_maps.search_type[table_name] == 'last':
                start_test_window = defaults.nem_data_model_start_time
                #start_test_window = '2018/01/01 00:00:00'
                start_time = datetime.strptime(start_test_window, '%Y/%m/%d %H:%M:%S')
                end_time = datetime.strptime('2018/01/01 00:00:00', '%Y/%m/%d %H:%M:%S')
                start_search = datetime.strptime(start_test_window, '%Y/%m/%d %H:%M:%S')
                data_tables = data_fetch_methods.dynamic_data_fetch_loop(
                    start_search=start_search, start_time=start_time,
                    end_time=end_time, table_name=table_name, raw_data_location='E:/raw_aemo_data',
                    select_columns=None, date_filter=None,
                    search_type='all')
                first_data_table = data_tables[35].loc[:, defaults.table_primary_keys[table_name]]
                last_data_table = data_tables[-1]
                comp = pd.merge(first_data_table, last_data_table, 'left', defaults.table_primary_keys[table_name])
                non_primary_col = [col for col in defaults.table_columns[table_name]
                                  if col not in defaults.table_primary_keys[table_name]][0]
                missing_from_last = comp[comp[non_primary_col].isnull()]
                self.assertEqual(False, missing_from_last.empty)







import unittest
from datetime import datetime
from nemosis import date_generators


class TestYearAndMonthGen(unittest.TestCase):
    def setUp(self):
        pass

    def test_two_times_not_at_edge_of_month_return_one_month(self):
        start_time = datetime.strptime('2017/01/02 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2017/01/03 00:00:00', '%Y/%m/%d %H:%M:%S')
        gen = date_generators.year_and_month_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], '2017')
        self.assertEqual(times[0][1], '01')
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(len(times), 1)


    def test_two_times_first_at_edge_of_month_return_month_before_and_month_of_times(self):
        start_time = datetime.strptime('2017/02/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2017/02/03 00:00:00', '%Y/%m/%d %H:%M:%S')
        gen = date_generators.year_and_month_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], '2017')
        self.assertEqual(times[0][1], '01')
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], '2017')
        self.assertEqual(times[1][1], '02')
        self.assertEqual(times[1][2], None)
        self.assertEqual(times[1][3], None)

    def test_two_times_first_at_edge_of_year_return_month_before_and_month_of_times(self):
        start_time = datetime.strptime('2017/01/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2017/01/03 00:00:00', '%Y/%m/%d %H:%M:%S')
        gen = date_generators.year_and_month_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], '2016')
        self.assertEqual(times[0][1], '12')
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], '2017')
        self.assertEqual(times[1][1], '01')
        self.assertEqual(times[1][2], None)
        self.assertEqual(times[1][3], None)

    def test_two_times_second_at_edge_of_month_returns_one_month(self):
        start_time = datetime.strptime('2017/01/05 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2017/01/31 00:00:00', '%Y/%m/%d %H:%M:%S')
        gen = date_generators.year_and_month_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], '2017')
        self.assertEqual(times[0][1], '01')
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(len(times), 1)

    def test_two_times_second_at_edge_of_year_returns_one_month(self):
        start_time = datetime.strptime('2017/12/02 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2017/12/31 00:00:00', '%Y/%m/%d %H:%M:%S')
        gen = date_generators.year_and_month_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], '2017')
        self.assertEqual(times[0][1], '12')
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(len(times), 1)

    def test_two_times_in_middle_of_jan_and_march_return_3_months(self):
        start_time = datetime.strptime('2017/01/05 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2017/03/24 00:00:00', '%Y/%m/%d %H:%M:%S')
        gen = date_generators.year_and_month_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], '2017')
        self.assertEqual(times[0][1], '01')
        self.assertEqual(times[0][2], None)
        self.assertEqual(times[0][3], None)
        self.assertEqual(times[1][0], '2017')
        self.assertEqual(times[1][1], '02')
        self.assertEqual(times[1][2], None)
        self.assertEqual(times[1][3], None)
        self.assertEqual(times[2][0], '2017')
        self.assertEqual(times[2][1], '03')
        self.assertEqual(times[2][2], None)
        self.assertEqual(times[2][3], None)
        self.assertEqual(len(times), 3)


class TestYearMonthDayIndexGen(unittest.TestCase):
    def setUp(self):
        pass

    def test_two_times_in_middle_of_adjacent_hours(self):
        # Whole of each hour should be returned.
        start_time = datetime.strptime('2013/01/05 12:20:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2013/01/05 13:45:00', '%Y/%m/%d %H:%M:%S')
        gen = date_generators.year_month_day_index_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], '2013')
        self.assertEqual(times[0][1], '01')
        self.assertEqual(times[0][2], '05')
        self.assertEqual(times[0][3], '1355')
        self.assertEqual(times[-1][0], '2013')
        self.assertEqual(times[-1][1], '01')
        self.assertEqual(times[-1][2], '05')
        self.assertEqual(times[-1][3], '1200')
        self.assertEqual(len(times), 24)

    def test_two_times_one_at_start_of_year_should_not_over_flow_to_previous_year(self):
        start_time = datetime.strptime('2013/01/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2013/01/01 01:00:00', '%Y/%m/%d %H:%M:%S')
        gen = date_generators.year_month_day_index_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], '2013')
        self.assertEqual(times[0][1], '01')
        self.assertEqual(times[0][2], '01')
        self.assertEqual(times[0][3], '0155')
        self.assertEqual(times[-1][0], '2013')
        self.assertEqual(times[-1][1], '01')
        self.assertEqual(times[-1][2], '01')
        self.assertEqual(times[-1][3], '0000')
        self.assertEqual(len(times), 24)

    def test_two_times_one_at_start_of_month_should_not_over_flow_to_previous_month(self):
        start_time = datetime.strptime('2013/02/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2013/02/01 01:00:00', '%Y/%m/%d %H:%M:%S')
        gen = date_generators.year_month_day_index_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], '2013')
        self.assertEqual(times[0][1], '02')
        self.assertEqual(times[0][2], '01')
        self.assertEqual(times[0][3], '0155')
        self.assertEqual(times[-1][0], '2013')
        self.assertEqual(times[-1][1], '02')
        self.assertEqual(times[-1][2], '01')
        self.assertEqual(times[-1][3], '0000')
        self.assertEqual(len(times), 24)

    def test_no_missing_values_in_a_week(self):
        start_time = datetime.strptime('2013/02/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        end_time = datetime.strptime('2013/02/07 00:00:00', '%Y/%m/%d %H:%M:%S')
        gen = date_generators.year_month_day_index_gen(start_time, end_time)
        times = [(year, month, day, index) for year, month, day, index in gen]
        self.assertEqual(times[0][0], '2013')
        self.assertEqual(times[0][1], '02')
        self.assertEqual(times[0][2], '01')
        self.assertEqual(times[0][3], '2355')
        self.assertEqual(times[-1][0], '2013')
        self.assertEqual(times[-1][1], '02')
        self.assertEqual(times[-1][2], '07')
        self.assertEqual(times[-1][3], '0000')
        self.assertEqual(len(times), 1740)

